@@ -5,24 +5,24 @@ from UM.Math.Vector import Vector
 
 
 class Ray:
-    def __init__(self, origin = Vector(), direction = Vector()):
+    def __init__(self, origin: Vector = Vector(), direction: Vector = Vector()) -> None:
         self._origin = origin
         self._direction = direction
         self._inverse_direction = 1.0 / direction
 
     @property
-    def origin(self):
+    def origin(self) -> Vector:
         return self._origin
 
     @property
-    def direction(self):
+    def direction(self) -> Vector:
         return self._direction
 
     @property
-    def inverseDirection(self):
+    def inverseDirection(self) -> Vector:
         return self._inverse_direction
 
-    def getPointAlongRay(self, distance):
+    def getPointAlongRay(self, distance: float) -> Vector:
         return self._origin + (self._direction * distance)
 
     def __repr__(self):
