@@ -13,21 +13,25 @@ if TYPE_CHECKING:
     from UM.Qt.QtApplication import QtApplication
 
 
-##  Central class for reading and writing meshes.
-#   This class is created by Application and handles reading and writing mesh files.
 class MeshFileHandler(FileHandler):
+    """Central class for reading and writing meshes.
+
+    This class is created by Application and handles reading and writing mesh files.
+    """
 
     def __init__(self, application: "QtApplication", writer_type: str = "mesh_writer", reader_type: str = "mesh_reader", parent: QObject = None) -> None:
         super().__init__(application, writer_type, reader_type, parent)
 
-    # Try to read the mesh_data from a file using a specified MeshReader.
-    # \param reader the MeshReader to read the file with.
-    # \param file_name The name of the mesh to load.
-    # \param kwargs Keyword arguments.
-    #               Possible values are:
-    #               - Center: True if the model should be centered around (0,0,0), False if it should be loaded as-is. Defaults to True.
-    # \returns MeshData if it was able to read the file, None otherwise.
     def readerRead(self, reader, file_name, **kwargs):
+        """Try to read the mesh_data from a file using a specified MeshReader.
+        :param reader: the MeshReader to read the file with.
+        :param file_name: The name of the mesh to load.
+        :param kwargs: Keyword arguments.
+        Possible values are:
+        - Center: True if the model should be centered around (0,0,0), False if it should be loaded as-is. Defaults to True.
+        :returns: MeshData if it was able to read the file, None otherwise.
+        """
+
         try:
             results = reader.read(file_name)
             if results is not None:
