@@ -11,7 +11,7 @@ from PyQt5.QtCore import Qt, QCoreApplication, QEvent, QUrl, pyqtProperty, pyqtS
 from UM.FlameProfiler import pyqtSlot
 from PyQt5.QtQml import QQmlApplicationEngine, QQmlComponent, QQmlContext, QQmlError
 from PyQt5.QtWidgets import QApplication, QSplashScreen, QMessageBox, QSystemTrayIcon
-from PyQt5.QtGui import QIcon, QPixmap, QFontMetrics
+from PyQt5.QtGui import QIcon, QPixmap, QFontMetrics, QSurfaceFormat
 from PyQt5.QtCore import QTimer
 
 from UM.Backend.Backend import Backend #For typing.
@@ -37,6 +37,8 @@ from UM.VersionUpgradeManager import VersionUpgradeManager
 from UM.View.GL.OpenGLContext import OpenGLContext
 from UM.Version import Version
 
+from UM.TaskManagement.HttpRequestManager import HttpRequestManager
+
 from UM.Operations.GroupedOperation import GroupedOperation #To clear the scene.
 from UM.Operations.RemoveSceneNodeOperation import RemoveSceneNodeOperation #To clear the scene.
 from UM.Scene.Iterator.DepthFirstIterator import DepthFirstIterator #To clear the scene.
@@ -64,12 +66,13 @@ if int(major) < 5 or (int(major) == 5 and int(minor) < 9):
     raise UnsupportedVersionError("This application requires at least PyQt 5.9.0")
 
 
-##  Application subclass that provides a Qt application object.
 @signalemitter
 class QtApplication(QApplication, Application):
+    """Application subclass that provides a Qt application object."""
+
     pluginsLoaded = Signal()
     applicationRunning = Signal()
-    
+
     def __init__(self, tray_icon_name: str = None, **kwargs) -> None:
         plugin_path = ""
         if sys.platform == "win32":
@@ -112,6 +115,12 @@ class QtApplication(QApplication, Application):
 
         self._configuration_error_message = None #type: Optional[ConfigurationErrorMessage]
 
+        self._http_network_request_manager = HttpRequestManager(parent = self)
+
+        #Metadata required for the file dialogues.
+        self.setOrganizationDomain("https://ultimaker.com/")
+        self.setOrganizationName("Ultimaker B.V.")
+
     def addCommandLineOptions(self) -> None:
         super().addCommandLineOptions()
         # This flag is used by QApplication. We don't process it.
@@ -120,6 +129,20 @@ class QtApplication(QApplication, Application):
 
     def initialize(self) -> None:
         super().initialize()
+
+        preferences = Application.getInstance().getPreferences()
+        preferences.addPreference("view/force_empty_shader_cache", False)
+        preferences.addPreference("view/opengl_version_detect", OpenGLContext.OpenGlVersionDetect.Autodetect)
+
+        # Read preferences here (upgrade won't work) to get:
+        #  - The language in use, so the splash window can be shown in the correct language.
+        #  - The OpenGL 'force' parameters.
+        try:
+            preferences_filename = Resources.getPath(Resources.Preferences, self._app_name + ".cfg")
+            self._preferences.readFromFile(preferences_filename)
+        except FileNotFoundError:
+            Logger.log("i", "Preferences file not found, ignore and use default language '%s'", self._default_language)
+
         # Initialize the package manager to remove and install scheduled packages.
         self._package_manager = self._package_manager_class(self, parent = self)
 
@@ -129,13 +152,21 @@ class QtApplication(QApplication, Application):
         # Remove this and you will get Windows 95 style for all widgets if you are using Qt 5.10+
         self.setStyle("fusion")
 
+        if preferences.getValue("view/force_empty_shader_cache"):
+            self.setAttribute(Qt.AA_DisableShaderDiskCache)
         self.setAttribute(Qt.AA_UseDesktopOpenGL)
-        major_version, minor_version, profile = OpenGLContext.detectBestOpenGLVersion()
+        if preferences.getValue("view/opengl_version_detect") != OpenGLContext.OpenGlVersionDetect.ForceModern:
+            major_version, minor_version, profile = OpenGLContext.detectBestOpenGLVersion(
+                preferences.getValue("view/opengl_version_detect") == OpenGLContext.OpenGlVersionDetect.ForceLegacy)
+        else:
+            Logger.info("Force 'modern' OpenGL (4.1 core) -- overrides 'force legacy opengl' preference.")
+            major_version, minor_version, profile = (4, 1, QSurfaceFormat.CoreProfile)
 
-        if major_version is None and minor_version is None and profile is None and not self.getIsHeadLess():
+        if major_version is None or minor_version is None or profile is None:
             Logger.log("e", "Startup failed because OpenGL version probing has failed: tried to create a 2.0 and 4.1 context. Exiting")
-            QMessageBox.critical(None, "Failed to probe OpenGL",
-                                 "Could not probe OpenGL. This program requires OpenGL 2.0 or higher. Please check your video card drivers.")
+            if not self.getIsHeadLess():
+                QMessageBox.critical(None, "Failed to probe OpenGL",
+                                     "Could not probe OpenGL. This program requires OpenGL 2.0 or higher. Please check your video card drivers.")
             sys.exit(1)
         else:
             opengl_version_str = OpenGLContext.versionAsText(major_version, minor_version, profile)
@@ -159,14 +190,6 @@ class QtApplication(QApplication, Application):
         self.showSplashMessage(i18n_catalog.i18nc("@info:progress", "Initializing package manager..."))
         self._package_manager.initialize()
 
-        # Read preferences here (upgrade won't work) to get the language in use, so the splash window can be shown in
-        # the correct language.
-        try:
-            preferences_filename = Resources.getPath(Resources.Preferences, self._app_name + ".cfg")
-            self._preferences.readFromFile(preferences_filename)
-        except FileNotFoundError:
-            Logger.log("i", "Preferences file not found, ignore and use default language '%s'", self._default_language)
-
         signal.signal(signal.SIGINT, signal.SIG_DFL)
         # This is done here as a lot of plugins require a correct gl context. If you want to change the framework,
         # these checks need to be done in your <framework>Application.py class __init__().
@@ -190,6 +213,7 @@ class QtApplication(QApplication, Application):
 
         # Load preferences again because before we have loaded the plugins, we don't have the upgrade routine for
         # the preferences file. Now that we have, load the preferences file again so it can be upgraded and loaded.
+        self.showSplashMessage(i18n_catalog.i18nc("@info:progress", "Loading preferences..."))
         try:
             preferences_filename = Resources.getPath(Resources.Preferences, self._app_name + ".cfg")
             with open(preferences_filename, "r", encoding = "utf-8") as f:
@@ -201,8 +225,8 @@ class QtApplication(QApplication, Application):
         except (FileNotFoundError, UnicodeDecodeError):
             Logger.log("i", "The preferences file cannot be found or it is corrupted, so we will use default values")
 
+        self.processEvents()
         # Force the configuration file to be written again since the list of plugins to remove maybe changed
-        self.showSplashMessage(i18n_catalog.i18nc("@info:progress", "Loading preferences..."))
         try:
             self._preferences_filename = Resources.getPath(Resources.Preferences, self._app_name + ".cfg")
             self._preferences.readFromFile(self._preferences_filename)
@@ -240,13 +264,17 @@ class QtApplication(QApplication, Application):
             # Initialize System tray icon and make it invisible because it is used only to show pop up messages
             self._tray_icon = None
             if self._tray_icon_name:
-                self._tray_icon = QIcon(Resources.getPath(Resources.Images, self._tray_icon_name))
-                self._tray_icon_widget = QSystemTrayIcon(self._tray_icon)
-                self._tray_icon_widget.setVisible(False)
+                try:
+                    self._tray_icon = QIcon(Resources.getPath(Resources.Images, self._tray_icon_name))
+                    self._tray_icon_widget = QSystemTrayIcon(self._tray_icon)
+                    self._tray_icon_widget.setVisible(False)
+                except FileNotFoundError:
+                    Logger.log("w", "Could not find the icon %s", self._tray_icon_name)
 
     def initializeEngine(self) -> None:
         # TODO: Document native/qml import trickery
         self._qml_engine = QQmlApplicationEngine(self)
+        self.processEvents()
         self._qml_engine.setOutputWarningsToStandardError(False)
         self._qml_engine.warnings.connect(self.__onQmlWarning)
 
@@ -257,11 +285,20 @@ class QtApplication(QApplication, Application):
             self._qml_engine.addImportPath(os.path.join(os.path.dirname(__file__), "qml"))
 
         self._qml_engine.rootContext().setContextProperty("QT_VERSION_STR", QT_VERSION_STR)
+        self.processEvents()
         self._qml_engine.rootContext().setContextProperty("screenScaleFactor", self._screenScaleFactor())
 
         self.registerObjects(self._qml_engine)
 
         Bindings.register()
+
+        # Preload theme. The theme will be loaded on first use, which will incur a ~0.1s freeze on the MainThread.
+        # Do it here, while the splash screen is shown. Also makes this freeze explicit and traceable.
+        self.getTheme()
+        self.processEvents()
+
+        i18n_catalog = i18nCatalog("uranium")
+        self.showSplashMessage(i18n_catalog.i18nc("@info:progress", "Loading UI..."))
         self._qml_engine.load(self._main_qml)
         self.engineCreatedSignal.emit()
 
@@ -339,7 +376,7 @@ class QtApplication(QApplication, Application):
     def exec_(self, *args: Any, **kwargs: Any) -> None:
         self.applicationRunning.emit()
         super().exec_(*args, **kwargs)
-        
+
     @pyqtSlot()
     def reloadQML(self) -> None:
         # only reload when it is a release build
@@ -395,7 +432,6 @@ class QtApplication(QApplication, Application):
             self._main_window = window
             if self._main_window is not None:
                 self._main_window.windowStateChanged.connect(self._onMainWindowStateChanged)
-
             self.mainWindowChanged.emit()
 
     def setVisible(self, visible: bool) -> None:
@@ -464,23 +500,30 @@ class QtApplication(QApplication, Application):
         else:
             return False
 
-    ##  Get the backend of the application (the program that does the heavy lifting).
-    #   The backend is also a QObject, which can be used from qml.
     @pyqtSlot(result = "QObject*")
     def getBackend(self) -> Backend:
+        """Get the backend of the application (the program that does the heavy lifting).
+
+        The backend is also a QObject, which can be used from qml.
+        """
+
         return self._backend
 
-    ##  Property used to expose the backend
-    #   It is made static as the backend is not supposed to change during runtime.
-    #   This makes the connection between backend and QML more reliable than the pyqtSlot above.
-    #   \returns Backend \type{Backend}
     @pyqtProperty("QVariant", constant = True)
     def backend(self) -> Backend:
+        """Property used to expose the backend
+
+        It is made static as the backend is not supposed to change during runtime.
+        This makes the connection between backend and QML more reliable than the pyqtSlot above.
+        :returns: Backend :type{Backend}
+        """
+
         return self.getBackend()
 
-    ## Create a class variable so we can manage the splash in the CrashHandler dialog when the Application instance
-    # is not yet created, e.g. when an error occurs during the initialization
     splash = None  # type: Optional[QSplashScreen]
+    """Create a class variable so we can manage the splash in the CrashHandler dialog when the Application instance
+    is not yet created, e.g. when an error occurs during the initialization
+    """
 
     def createSplash(self) -> None:
         if not self.getIsHeadLess():
@@ -493,30 +536,35 @@ class QtApplication(QApplication, Application):
                     QtApplication.splash.show()
                     self.processEvents()
 
-    ##  Display text on the splash screen.
     def showSplashMessage(self, message: str) -> None:
+        """Display text on the splash screen."""
+
         if not QtApplication.splash:
             self.createSplash()
-        
+
         if QtApplication.splash:
-            QtApplication.splash.showMessage(message, Qt.AlignHCenter | Qt.AlignVCenter)
-            self.processEvents()
+            self.processEvents()  # Process events from previous loading phase before updating the message
+            QtApplication.splash.showMessage(message, Qt.AlignHCenter | Qt.AlignVCenter)  # Now update the message
+            self.processEvents()  # And make sure it is immediately visible
         elif self.getIsHeadLess():
             Logger.log("d", message)
 
-    ##  Close the splash screen after the application has started.
     def closeSplash(self) -> None:
+        """Close the splash screen after the application has started."""
+
         if QtApplication.splash:
             QtApplication.splash.close()
             QtApplication.splash = None
 
-    ## Create a QML component from a qml file.
-    #  \param qml_file_path: The absolute file path to the root qml file.
-    #  \param context_properties: Optional dictionary containing the properties that will be set on the context of the
-    #                              qml instance before creation.
-    #  \return None in case the creation failed (qml error), else it returns the qml instance.
-    #  \note If the creation fails, this function will ensure any errors are logged to the logging service.
     def createQmlComponent(self, qml_file_path: str, context_properties: Dict[str, "QObject"] = None) -> Optional["QObject"]:
+        """Create a QML component from a qml file.
+        :param qml_file_path:: The absolute file path to the root qml file.
+        :param context_properties:: Optional dictionary containing the properties that will be set on the context of the
+        qml instance before creation.
+        :return: None in case the creation failed (qml error), else it returns the qml instance.
+        :note If the creation fails, this function will ensure any errors are logged to the logging service.
+        """
+
         if self._qml_engine is None: # Protect in case the engine was not initialized yet
             return None
         path = QUrl.fromLocalFile(qml_file_path)
@@ -536,14 +584,23 @@ class QtApplication(QApplication, Application):
         result.attached_context = result_context
         return result
 
-    ##  Delete all nodes containing mesh data in the scene.
-    #   \param only_selectable. Set this to False to delete objects from all build plates
     @pyqtSlot()
     def deleteAll(self, only_selectable = True) -> None:
+        """Delete all nodes containing mesh data in the scene.
+        :param only_selectable:. Set this to False to delete objects from all build plates
+        """
+
         self.getController().deleteAllNodesWithMeshData(only_selectable)
 
-    ##  Get the MeshFileHandler of this application.
+    @pyqtSlot()
+    def resetWorkspace(self) -> None:
+        self._workspace_metadata_storage.clear()
+        self.deleteAll()
+        self.workspaceLoaded.emit("")
+
     def getMeshFileHandler(self) -> MeshFileHandler:
+        """Get the MeshFileHandler of this application."""
+
         return self._mesh_file_handler
 
     def getWorkspaceFileHandler(self) -> WorkspaceFileHandler:
@@ -553,12 +610,17 @@ class QtApplication(QApplication, Application):
     def getPackageManager(self) -> PackageManager:
         return self._package_manager
 
-    ##  Gets the instance of this application.
-    #
-    #   This is just to further specify the type of Application.getInstance().
-    #   \return The instance of this application.
+    def getHttpRequestManager(self) -> "HttpRequestManager":
+        return self._http_network_request_manager
+
     @classmethod
     def getInstance(cls, *args, **kwargs) -> "QtApplication":
+        """Gets the instance of this application.
+
+        This is just to further specify the type of Application.getInstance().
+        :return: The instance of this application.
+        """
+
         return cast(QtApplication, super().getInstance(**kwargs))
 
     def _createSplashScreen(self) -> QSplashScreen:
@@ -576,11 +638,17 @@ class QtApplication(QApplication, Application):
             fontPixelRatio = int(fontPixelRatio * 4) / 4
             return fontPixelRatio
 
+    @pyqtProperty(str, constant=True)
+    def applicationDisplayName(self) -> str:
+        return self.getApplicationDisplayName()
+
 
-##  Internal.
-#
-#   Wrapper around a FunctionEvent object to make Qt handle the event properly.
 class _QtFunctionEvent(QEvent):
+    """Internal.
+
+    Wrapper around a FunctionEvent object to make Qt handle the event properly.
+    """
+
     QtFunctionEvent = QEvent.User + 1
 
     def __init__(self, fevent: QEvent) -> None:
