@@ -1,11 +1,10 @@
-# Copyright (c) 2018 Ultimaker B.V.
+# Copyright (c) 2020 Ultimaker B.V.
 # Uranium is released under the terms of the LGPLv3 or higher.
 
 import numpy
 from PyQt5.QtGui import QColor, QOpenGLBuffer, QOpenGLVertexArrayObject
-from typing import List, Optional, Tuple
+from typing import List, Tuple, Dict
 
-from UM.Application import Application
 import UM.Qt.QtApplication
 from UM.View.Renderer import Renderer
 from UM.Math.Vector import Vector
@@ -25,9 +24,6 @@ from UM.Logger import Logger
 
 MYPY = False
 if MYPY:
-    from UM.Controller import Controller
-    from UM.Scene.Scene import Scene
-    from UM.Scene.Camera import Camera
     from UM.Scene.SceneNode import SceneNode
     from UM.View.RenderPass import RenderPass
     from UM.View.GL.ShaderProgram import ShaderProgram
@@ -37,15 +33,13 @@ vertexBufferProperty = "__qtgl2_vertex_buffer"
 indexBufferProperty = "__qtgl2_index_buffer"
 
 
-##  A Renderer implementation using PyQt's OpenGL implementation to render.
 @signalemitter
 class QtRenderer(Renderer):
+    """A Renderer implementation using PyQt's OpenGL implementation to render."""
+
     def __init__(self) -> None:
         super().__init__()
 
-        self._controller = Application.getInstance().getController()  # type: Controller
-        self._scene = self._controller.getScene()  # type: Scene
-
         self._initialized = False  # type: bool
 
         self._light_position = Vector(0, 0, 0)  # type: Vector
@@ -56,32 +50,44 @@ class QtRenderer(Renderer):
         self._window_height = 0  # type: int
 
         self._batches = []  # type: List[RenderBatch]
-
+        self._named_batches = {}  # type: Dict[str, RenderBatch]
         self._quad_buffer = None  # type: QOpenGLBuffer
 
-        self._camera = None  # type: Optional[Camera]
-
     initialized = Signal()
 
-    ##  Get an integer multiplier that can be used to correct for screen DPI.
-    def getPixelMultiplier(self) -> int:
+    @staticmethod
+    def getPixelMultiplier() -> int:
+        """Get an integer multiplier that can be used to correct for screen DPI."""
+
         # Standard assumption for screen pixel density is 96 DPI. We use that as baseline to get
         # a multiplication factor we can use for screens > 96 DPI.
         return round(UM.Qt.QtApplication.QtApplication.getInstance().primaryScreen().physicalDotsPerInch() / 96.0)
 
-    ##  Get the list of render batches.
     def getBatches(self) -> List[RenderBatch]:
+        """Get the list of render batches."""
+
         return self._batches
 
-    ##  Overridden from Renderer.
-    #
-    #   This makes sure the added render pass has the right size.
+    def addRenderBatch(self, render_batch, name = ""):
+        self._batches.append(render_batch)
+        if name:
+            self._named_batches[name] = render_batch
+
+    def getNamedBatch(self, name):
+        return self._named_batches.get(name)
+
     def addRenderPass(self, render_pass: "RenderPass") -> None:
+        """Overridden from Renderer.
+
+        This makes sure the added render pass has the right size.
+        """
+
         super().addRenderPass(render_pass)
         render_pass.setSize(self._viewport_width, self._viewport_height)
 
-    ##  Set background color used when rendering.
     def setBackgroundColor(self, color: QColor) -> None:
+        """Set background color used when rendering."""
+
         self._background_color = color
 
     def getViewportWidth(self) -> int:
@@ -90,30 +96,36 @@ class QtRenderer(Renderer):
     def getViewportHeight(self) -> int:
         return self._viewport_height
 
-    ##  Set the viewport size.
-    #
-    #   \param width The new width of the viewport.
-    #   \param height The new height of the viewport.
     def setViewportSize(self, width: int, height: int) -> None:
+        """Set the viewport size.
+
+        :param width: The new width of the viewport.
+        :param height: The new height of the viewport.
+        """
+
         self._viewport_width = width
         self._viewport_height = height
 
         for render_pass in self._render_passes:
             render_pass.setSize(width, height)
 
-    ##  Set the window size.
     def setWindowSize(self, width: int, height: int) -> None:
+        """Set the window size."""
+
         self._window_width = width
         self._window_height = height
 
-    ##  Get the window size.
-    #
-    #   \return A tuple of (window_width, window_height)
     def getWindowSize(self) -> Tuple[int, int]:
+        """Get the window size.
+
+        :return: A tuple of (window_width, window_height)
+        """
+
         return self._window_width, self._window_height
 
-    ##  Overrides Renderer::beginRendering()
     def beginRendering(self) -> None:
+        """Overrides Renderer::beginRendering()"""
+
         if not self._initialized:
             self._initialize()
 
@@ -122,8 +134,15 @@ class QtRenderer(Renderer):
         self._gl.glClear(self._gl.GL_COLOR_BUFFER_BIT | self._gl.GL_DEPTH_BUFFER_BIT)
         self._gl.glClearColor(0.0, 0.0, 0.0, 0.0)
 
-    ##  Overrides Renderer::queueNode()
     def queueNode(self, node: "SceneNode", **kwargs) -> None:
+        """Overrides Renderer::queueNode()"""
+        batch = self.createRenderBatch(**kwargs)
+
+        batch.addItem(node.getWorldTransformation(copy = False), kwargs.get("mesh", node.getMeshData()), kwargs.pop("uniforms", None), normal_transformation=node.getCachedNormalMatrix())
+
+        self._batches.append(batch)
+
+    def createRenderBatch(self, **kwargs):
         type = kwargs.pop("type", RenderBatch.RenderType.Solid)
         if kwargs.pop("transparent", False):
             type = RenderBatch.RenderType.Transparent
@@ -131,14 +150,11 @@ class QtRenderer(Renderer):
             type = RenderBatch.RenderType.Overlay
 
         shader = kwargs.pop("shader", self._default_material)
-        batch = RenderBatch(shader, type = type, **kwargs)
+        return RenderBatch(shader, type=type, **kwargs)
 
-        batch.addItem(node.getWorldTransformation(), kwargs.get("mesh", node.getMeshData()), kwargs.pop("uniforms", None))
-
-        self._batches.append(batch)
-
-    ##  Overrides Renderer::render()
     def render(self) -> None:
+        """Overrides Renderer::render()"""
+
         self._batches.sort()
 
         for render_pass in self.getRenderPasses():
@@ -146,15 +162,28 @@ class QtRenderer(Renderer):
             self._gl.glViewport(0, 0, width, height)
             render_pass.render()
 
-    ##  Overrides Renderer::endRendering()
+    def reRenderLast(self):
+        """Sometimes not an *entire* new render is required (QML is updated, but the scene did not).
+        In that case we ask the composite pass (which must be the last render pass) to render (instead of re-rendering
+        all the passes & the views.
+        """
+
+        self.beginRendering() # First ensure that the viewport is set correctly.
+        self.getRenderPasses()[-1].render()
+
     def endRendering(self) -> None:
+        """Overrides Renderer::endRendering()"""
+
         self._batches.clear()
+        self._named_batches.clear()
 
-    ##  Render a full screen quad (rectangle).
-    #
-    #   The function is used to draw render results on.
-    #   \param shader The shader to use when rendering.
     def renderFullScreenQuad(self, shader: "ShaderProgram") -> None:
+        """Render a full screen quad (rectangle).
+
+        The function is used to draw render results on.
+        :param shader: The shader to use when rendering.
+        """
+
         self._gl.glDisable(self._gl.GL_DEPTH_TEST)
         self._gl.glDisable(self._gl.GL_BLEND)
 
@@ -185,9 +214,9 @@ class QtRenderer(Renderer):
 
         self._default_material = OpenGL.getInstance().createShaderProgram(Resources.getPath(Resources.Shaders, "default.shader")) #type: ShaderProgram
 
-        self._render_passes.add(DefaultPass(self._viewport_width, self._viewport_height))
-        self._render_passes.add(SelectionPass(self._viewport_width, self._viewport_height))
-        self._render_passes.add(CompositePass(self._viewport_width, self._viewport_height))
+        self.addRenderPass(DefaultPass(self._viewport_width, self._viewport_height))
+        self.addRenderPass(SelectionPass(self._viewport_width, self._viewport_height))
+        self.addRenderPass(CompositePass(self._viewport_width, self._viewport_height))
 
         buffer = QOpenGLBuffer(QOpenGLBuffer.VertexBuffer)
         buffer.create()
