@@ -1,5 +1,6 @@
 # Copyright (c) 2017 Ultimaker B.V.
 # Uranium is released under the terms of the LGPLv3 or higher.
+from typing import Optional, List, Set, Any
 
 from PyQt5.QtCore import QObject, QTimer, pyqtProperty, pyqtSignal
 from PyQt5.QtQml import QQmlPropertyMap
@@ -7,120 +8,146 @@ from UM.FlameProfiler import pyqtSlot
 
 from UM.Logger import Logger
 from UM.Application import Application
+from UM.Settings.ContainerStack import ContainerStack
 from UM.Settings.SettingFunction import SettingFunction
 from UM.Settings.ContainerRegistry import ContainerRegistry
 from UM.Settings.DefinitionContainer import DefinitionContainer
 from UM.Settings.InstanceContainer import InstanceContainer
-from UM.Settings.Interfaces import PropertyEvaluationContext
+from UM.Settings.Interfaces import PropertyEvaluationContext, ContainerInterface
 from UM.Settings.SettingInstance import InstanceState
 from UM.Settings.SettingRelation import RelationType
 from UM.Settings.SettingDefinition import SettingDefinition
+from UM.Settings.Validator import Validator
+
 
-##  This class provides the value and change notifications for the properties of a single setting
-#
-#   Since setting values and other properties are provided by a stack, we need some way to
-#   query the stack from QML to provide us with those values. This class takes care of that.
-#
-#   This class provides the property values through QObject dynamic properties so that they
-#   are available from QML.
 class SettingPropertyProvider(QObject):
-    def __init__(self, parent = None, *args, **kwargs):
-        super().__init__(parent = parent, *args, **kwargs)
+    """This class provides the value and change notifications for the properties of a single setting
+
+    Since setting values and other properties are provided by a stack, we need some way to
+    query the stack from QML to provide us with those values. This class takes care of that.
+
+    This class provides the property values through QObject dynamic properties so that they
+    are available from QML.
+    """
+
+    def __init__(self, parent = None) -> None:
+        super().__init__(parent = parent)
 
         self._property_map = QQmlPropertyMap(self)
 
-        self._stack_id = ""
-        self._stack = None
+        self._stack = None  # type: Optional[ContainerStack]
         self._key = ""
-        self._relations = set()
-        self._watched_properties = []
+        self._relations = set()  # type: Set[str]
+        self._watched_properties = []  # type: List[str]
         self._store_index = 0
-        self._value_used = None
-        self._stack_levels = []
+        self._value_used = None  # type: Optional[bool]
+        self._stack_levels = []  # type: List[int]
         self._remove_unused_value = True
-        self._validator = None
+        self._validator = None  # type: Optional[Validator]
 
-        self._update_timer = QTimer()
+        self._update_timer = QTimer(self)
         self._update_timer.setInterval(100)
         self._update_timer.setSingleShot(True)
         self._update_timer.timeout.connect(self._update)
 
         self.storeIndexChanged.connect(self._storeIndexChanged)
 
-    ##  Set the containerStackId property.
-    def setContainerStackId(self, stack_id):
-        if stack_id == self._stack_id:
-            return # No change.
-
-        self._stack_id = stack_id
+    def setContainerStack(self, stack: Optional[ContainerStack]) -> None:
+        if self._stack == stack:
+            return  # Nothing to do, attempting to set stack to the same value.
 
         if self._stack:
             self._stack.propertiesChanged.disconnect(self._onPropertiesChanged)
             self._stack.containersChanged.disconnect(self._containersChanged)
 
-        if self._stack_id:
-            if self._stack_id == "global":
-                self._stack = Application.getInstance().getGlobalContainerStack()
-            else:
-                stacks = ContainerRegistry.getInstance().findContainerStacks(id = self._stack_id)
-                if stacks:
-                    self._stack = stacks[0]
+        self._stack = stack
 
-            if self._stack:
-                self._stack.propertiesChanged.connect(self._onPropertiesChanged)
-                self._stack.containersChanged.connect(self._containersChanged)
-        else:
-            self._stack = None
+        if self._stack:
+            self._stack.propertiesChanged.connect(self._onPropertiesChanged)
+            self._stack.containersChanged.connect(self._containersChanged)
 
         self._validator = None
-        self._update()
-        self.containerStackIdChanged.emit()
+        self._updateDelayed()
+        self.containerStackChanged.emit()
+
+    def setContainerStackId(self, stack_id: str) -> None:
+        """Set the containerStackId property."""
+
+        if stack_id == self.containerStackId:
+            return  # No change.
+
+        if stack_id:
+            if stack_id == "global":
+                self.setContainerStack(Application.getInstance().getGlobalContainerStack())
+            else:
+                stacks = ContainerRegistry.getInstance().findContainerStacks(id = stack_id)
+                if stacks:
+                    self.setContainerStack(stacks[0])
+        else:
+            self.setContainerStack(None)
 
-    ##  Emitted when the containerStackId property changes.
     containerStackIdChanged = pyqtSignal()
-    ##  The ID of the container stack we should query for property values.
+    """Emitted when the containerStackId property changes."""
+
     @pyqtProperty(str, fset = setContainerStackId, notify = containerStackIdChanged)
-    def containerStackId(self):
-        return self._stack_id
+    def containerStackId(self) -> str:
+        """The ID of the container stack we should query for property values."""
+
+        if self._stack:
+            return self._stack.id
+
+        return ""
+
+    containerStackChanged = pyqtSignal()
+
+    @pyqtProperty(QObject, fset=setContainerStack, notify=containerStackChanged)
+    def containerStack(self) -> Optional[ContainerInterface]:
+        return self._stack
 
     removeUnusedValueChanged = pyqtSignal()
 
-    def setRemoveUnusedValue(self, remove_unused_value):
+    def setRemoveUnusedValue(self, remove_unused_value: bool) -> None:
         if self._remove_unused_value != remove_unused_value:
             self._remove_unused_value = remove_unused_value
             self.removeUnusedValueChanged.emit()
 
     @pyqtProperty(bool, fset = setRemoveUnusedValue, notify = removeUnusedValueChanged)
-    def removeUnusedValue(self):
+    def removeUnusedValue(self) -> bool:
         return self._remove_unused_value
 
-    ##  Set the watchedProperties property.
-    def setWatchedProperties(self, properties):
+    def setWatchedProperties(self, properties: List[str]) -> None:
+        """Set the watchedProperties property."""
+
         if properties != self._watched_properties:
             self._watched_properties = properties
-            self._update()
+            self._updateDelayed()
             self.watchedPropertiesChanged.emit()
 
-    ##  Emitted when the watchedProperties property changes.
     watchedPropertiesChanged = pyqtSignal()
-    ##  A list of property names that should be watched for changes.
-    @pyqtProperty("QVariantList", fset = setWatchedProperties, notify = watchedPropertiesChanged)
-    def watchedProperties(self):
+    """Emitted when the watchedProperties property changes."""
+
+    @pyqtProperty("QStringList", fset = setWatchedProperties, notify = watchedPropertiesChanged)
+    def watchedProperties(self) -> List[str]:
+        """A list of property names that should be watched for changes."""
+
         return self._watched_properties
 
-    ##  Set the key property.
-    def setKey(self, key):
+    def setKey(self, key: str) -> None:
+        """Set the key property."""
+
         if key != self._key:
             self._key = key
             self._validator = None
-            self._update()
+            self._updateDelayed()
             self.keyChanged.emit()
 
-    ##  Emitted when the key property changes.
     keyChanged = pyqtSignal()
-    ##  The key of the setting that we should provide property values for.
+    """Emitted when the key property changes."""
+
     @pyqtProperty(str, fset = setKey, notify = keyChanged)
     def key(self):
+        """The key of the setting that we should provide property values for."""
+
         return self._key
 
     propertiesChanged = pyqtSignal()
@@ -144,20 +171,23 @@ class SettingPropertyProvider(QObject):
 
     stackLevelChanged = pyqtSignal()
 
-    ##  At what levels in the stack does the value(s) for this setting occur?
     @pyqtProperty("QVariantList", notify = stackLevelChanged)
     def stackLevels(self):
+        """At what levels in the stack does the value(s) for this setting occur?"""
+
         if not self._stack:
             return [-1]
         return self._stack_levels
 
-    ##  Set the value of a property.
-    #
-    #   \param stack_index At which level in the stack should this property be set?
-    #   \param property_name The name of the property to set.
-    #   \param property_value The value of the property to set.
     @pyqtSlot(str, "QVariant")
     def setPropertyValue(self, property_name, property_value):
+        """Set the value of a property.
+
+        :param stack_index: At which level in the stack should this property be set?
+        :param property_name: The name of the property to set.
+        :param property_value: The value of the property to set.
+        """
+
         if not self._stack or not self._key:
             return
 
@@ -200,13 +230,16 @@ class SettingPropertyProvider(QObject):
 
         container.setProperty(self._key, property_name, property_value)
 
-    ##  Manually request the value of a property.
-    #   The most notable difference with the properties is that you have more control over at what point in the stack
-    #   you want the setting to be retrieved (instead of always taking the top one)
-    #   \param property_name The name of the property to get the value from.
-    #   \param stack_level the index of the container to get the value from.
     @pyqtSlot(str, int, result = "QVariant")
-    def getPropertyValue(self, property_name, stack_level):
+    def getPropertyValue(self, property_name: str, stack_level: int) -> Any:
+        """Manually request the value of a property.
+        The most notable difference with the properties is that you have more control over at what point in the stack
+        you want the setting to be retrieved (instead of always taking the top one)
+
+        :param property_name: The name of the property to get the value from.
+        :param stack_level: the index of the container to get the value from.
+        """
+
         try:
             # Because we continue to count if there are multiple linked stacks, we need to check what stack is targeted
             current_stack = self._stack
@@ -214,7 +247,7 @@ class SettingPropertyProvider(QObject):
                 num_containers = len(current_stack.getContainers())
                 if stack_level >= num_containers:
                     stack_level -= num_containers
-                    current_stack = self._stack.getNextStack()
+                    current_stack = current_stack.getNextStack()
                 else:
                     break  # Found the right stack
 
@@ -228,14 +261,18 @@ class SettingPropertyProvider(QObject):
             return None
         return value
 
+    @pyqtSlot(str, result = str)
+    def getPropertyValueAsString(self, property_name: str) -> str:
+        return self._getPropertyValue(property_name)
+
     @pyqtSlot(int)
-    def removeFromContainer(self, index):
+    def removeFromContainer(self, index: int) -> None:
         current_stack = self._stack
         while current_stack:
             num_containers = len(current_stack.getContainers())
             if index >= num_containers:
                 index -= num_containers
-                current_stack = self._stack.getNextStack()
+                current_stack = current_stack.getNextStack()
             else:
                 break  # Found the right stack
 
@@ -252,7 +289,7 @@ class SettingPropertyProvider(QObject):
 
     isValueUsedChanged = pyqtSignal()
     @pyqtProperty(bool, notify = isValueUsedChanged)
-    def isValueUsed(self):
+    def isValueUsed(self) -> bool:
         if self._value_used is not None:
             return self._value_used
         if not self._stack:
@@ -272,21 +309,28 @@ class SettingPropertyProvider(QObject):
 
             if self._stack.getProperty(key, "state") != InstanceState.User:
                 value_used_count += 1
+                break
 
             # If the setting has a formula the value is still used.
             if isinstance(self._stack.getRawProperty(key, "value"), SettingFunction):
                 value_used_count += 1
+                break
 
         self._value_used = relation_count == 0 or (relation_count > 0 and value_used_count != 0)
         return self._value_used
 
-    # protected:
-
-    def _onPropertiesChanged(self, key, property_names):
+    def _onPropertiesChanged(self, key: str, property_names: List[str]) -> None:
         if key != self._key:
             if key in self._relations:
                 self._value_used = None
-                self.isValueUsedChanged.emit()
+                try:
+                    self.isValueUsedChanged.emit()
+                except RuntimeError:
+                    # QtObject has been destroyed, no need to handle the signals anymore.
+                    # This can happen when the QtObject in C++ has been destroyed, but the python object hasn't quite
+                    # caught on yet. Once we call any signals, it will cause a runtimeError since all the underlying
+                    # logic to emit pyqtSignals is gone.
+                    return
             return
 
         has_values_changed = False
@@ -295,11 +339,25 @@ class SettingPropertyProvider(QObject):
                 continue
 
             has_values_changed = True
-            self._property_map.insert(property_name, self._getPropertyValue(property_name))
+            try:
+                self._property_map.insert(property_name, self._getPropertyValue(property_name))
+            except RuntimeError:
+                # QtObject has been destroyed, no need to handle the signals anymore.
+                # This can happen when the QtObject in C++ has been destroyed, but the python object hasn't quite
+                # caught on yet. Once we call any signals, it will cause a runtimeError since all the underlying
+                # logic to emit pyqtSignals is gone.
+                return
 
         self._updateStackLevels()
         if has_values_changed:
-            self.propertiesChanged.emit()
+            try:
+                self.propertiesChanged.emit()
+            except RuntimeError:
+                # QtObject has been destroyed, no need to handle the signals anymore.
+                # This can happen when the QtObject in C++ has been destroyed, but the python object hasn't quite
+                # caught on yet. Once we call any signals, it will cause a runtimeError since all the underlying
+                # logic to emit pyqtSignals is gone.
+                return
 
     def _update(self, container = None):
         if not self._stack or not self._watched_properties or not self._key:
@@ -314,22 +372,30 @@ class SettingPropertyProvider(QObject):
         for property_name in self._watched_properties:
             self._property_map.insert(property_name, self._getPropertyValue(property_name))
 
+        # Notify that the properties have been changed.Kewl
+        self.propertiesChanged.emit()
+
         # Force update of value_used
         self._value_used = None
         self.isValueUsedChanged.emit()
 
-    def _updateDelayed(self, container = None):
-        self._update_timer.start()
+    def _updateDelayed(self, container = None) -> None:
+        try:
+            self._update_timer.start()
+        except RuntimeError:
+            # Sometimes the python object is not yet deleted, but the wrapped part is already gone.
+            # In that case there is nothing else to do but ignore this.
+            pass
 
     def _containersChanged(self, container = None):
-        self._updateDelayed(container = container)
+        self._updateDelayed()
 
-    def _storeIndexChanged(self, container = None):
-        self._updateDelayed(container = container)
+    def _storeIndexChanged(self):
+        self._updateDelayed()
+
+    def _updateStackLevels(self) -> None:
+        """Updates the self._stack_levels field, which indicates at which levels in the stack the property is set."""
 
-    ##  Updates the self._stack_levels field, which indicates at which levels in
-    #   the stack the property is set.
-    def _updateStackLevels(self):
         levels = []
         # Start looking at the stack this provider is attached to.
         current_stack = self._stack
