@@ -2,11 +2,13 @@
 # Uranium is released under the terms of the LGPLv3 or higher.
 
 
-##  Convert a value to a boolean
-#
-#   \param \type{bool|str|int} any value.
-#   \return \type{bool}
 def parseBool(value):
+    """Convert a value to a boolean
+
+    :param :type{bool|str|int} any value.
+    :return: :type{bool}
+    """
+
     return value in [True, "True", "true", "Yes", "yes", 1]
 
 
