@@ -15,20 +15,21 @@ if MYPY:
     from UM.View.GL.ShaderProgram import ShaderProgram
 
 
-##  A RenderPass subclass providing the final composition render.
-#
-#   This render pass uses the other render passes to render a final composited image.
-#   By default, this consists of the output of the default pass, with an outline
-#   rendered on top of it using a convolution filter.
-#
-#   You can use setCompositeShader() to override the shader used for the composition.
-#   Additionally, setLayerBindings() can be used to set layer bindings, that is set,
-#   which layer is bound to which texture unit.
-#
-#   \note The CompositePass should always be last in the Renderer's rendering order.
-#   Therefore, when subclassing RenderPass make sure to use a priority lower than
-#   RenderPass.MaximumPriority.
 class CompositePass(RenderPass):
+    """A RenderPass subclass providing the final composition render.
+
+    This render pass uses the other render passes to render a final composited image.
+    By default, this consists of the output of the default pass, with an outline
+    rendered on top of it using a convolution filter.
+
+    You can use setCompositeShader() to override the shader used for the composition.
+    Additionally, setLayerBindings() can be used to set layer bindings, that is set,
+    which layer is bound to which texture unit.
+
+    :note The CompositePass should always be last in the Renderer's rendering order.
+    Therefore, when subclassing RenderPass make sure to use a priority lower than
+    RenderPass.MaximumPriority.
+    """
     def __init__(self, width, height):
         super().__init__("composite", width, height, RenderPass.MaximumPriority)
 
@@ -42,31 +43,32 @@ class CompositePass(RenderPass):
 
         self._layer_bindings = ["default", "selection"]
 
-    ##  Get the shader currently used for compositing.
     def getCompositeShader(self) -> "ShaderProgram":
+        """Get the shader currently used for compositing."""
         return self._shader
 
-    ##  Set the shader to use for compositing.
     def setCompositeShader(self, shader: "ShaderProgram") -> None:
+        """Set the shader to use for compositing."""
         self._shader = shader
 
-    ##  Get the current layer bindings.
     def getLayerBindings(self) -> List[str]:
+        """Get the current layer bindings."""
         return self._layer_bindings
 
-    ##  Set the layer bindings to use.
-    #
-    #   This should be a list of RenderPass names. The passes will be bound
-    #   to different texture units in the order specified. By default, the output
-    #   of the "default" RenderPass is bound to texture unit 0 and the output of
-    #   the "selection" RenderPass is bound to texture unit 1.
-    #
-    #   \param bindings The list of layer bindings to use.
     def setLayerBindings(self, bindings: List[str]) -> None:
+        """Set the layer bindings to use.
+
+        This should be a list of RenderPass names. The passes will be bound
+        to different texture units in the order specified. By default, the output
+        of the "default" RenderPass is bound to texture unit 0 and the output of
+        the "selection" RenderPass is bound to texture unit 1.
+
+        :param bindings: The list of layer bindings to use.
+        """
         self._layer_bindings = bindings
 
-    ##  Perform the actual rendering of the render pass.
     def render(self) -> None:
+        """Perform the actual rendering of the render pass."""
         self._shader.bind()
 
         outline_size = 2.0
